# coding=utf-8
"""
Koskya XML general Utilities Module
contains the Kosakya XML general Utilities Classes
"""
from collections.abc import Iterator

from lxml import etree

from typing import Any, TypeAlias
TyDocInfo: TypeAlias = etree.DocInfo
TyTree: TypeAlias = etree._Element

TyAny = Any
TyArr = list[Any]
TyBytes = bytes
TyDic = dict[Any, Any]
TyAoB = list[TyBytes]
TyAoD = list[TyDic]
TyAoX = list[str]
TyDoA = dict[Any, TyArr]
TyDoD = dict[Any, TyDic]
TyIterAny = Iterator[Any]
TyPath = str
TyToD = tuple[TyDic, ...]

TnAny = None | Any
TnArr = None | TyArr
TnBool = None | bool
TnDic = None | TyDic


class Xml2DicApc:

    # @staticmethod
    # def dic_normalize(dic):
    #     return {k: v[0] if len(v) == 1 else v for k, v in dic.items()}

    @staticmethod
    def sh_v(v, child, arr_child) -> str | Any | TyDic:
        a_key_to_agg: TyArr = ['Attribute', 'Step']
        if child.tag == etree.Comment:
            text: str = child.text.strip()
            return text
        if child.tag not in a_key_to_agg:
            return v
        if not child.attrib and not child.text:
            return v
        if child.attrib:
            dic: TyDic = {f'@{k}': v for k, v in child.attrib.items()}
        else:
            dic = {}
        if child.text:
            text = child.text.strip()
            if len(arr_child) > 0 or child.attrib:
                if text:
                    if not dic:
                        dic = {}
                    dic['#text'] = text
            return text
        if not dic:
            dic = {}
        return dic

    @classmethod
    def tree2dic(cls, tree: TyTree) -> TyDic:
        obj: TyDic = {tree.tag: [] if tree.attrib else None}
        a_child: TyArr = list(tree)
        if len(a_child) > 0:
            aod: Iterator[TyDic] = map(cls.tree2dic, a_child)
            dd: TyDic = {}
            a_tree: TyArr = []
            for ix, dic in enumerate(aod):
                child = a_child[ix]
                if child.tag == etree.Comment:
                    continue
                for k, v in dic.items():
                    v = cls.sh_v(v, child, a_child)
                    a_key_to_agg = ['Attribute', 'Step']
                    if k in a_key_to_agg:
                        if k not in dd:
                            dd[k] = []
                        dd[k].append(v)
                    else:
                        if len(dd) > 0:
                            a_tree.append(dd)
                            dd = {}
                        _dic = {}
                        _dic[k] = v
                        a_tree.append(_dic)
            if len(dd) > 0:
                a_tree.append(dd)
                dd = {}
            obj = {tree.tag: a_tree}
            # obj = {tree.tag: cls.dic_normalize(dd)}
        # obj = set_obj(obj, tree, a_child)
        return obj

    @classmethod
    def mig(cls, str, **kwargs) -> TyDoD :
        """
        migrate xml string to etree and etree to dictionary
        """
        tree: TyTree = etree.fromstring(str)
        # Iterate through all XML elements
        for elem in tree.getiterator():
            # Skip comments and processing instructions,
            # because they do not have names
            if not (
                isinstance(elem, etree._Comment)
                or isinstance(elem, etree._ProcessingInstruction)
            ):
                # Remove a namespace URI in the element's name
                elem.tag = etree.QName(elem).localname
        etree.cleanup_namespaces(tree)
        dic = cls.tree2dic(tree)
        return {tree.tag: dic}



# coding=utf-8
"""
Koskya XML general Utilities Module
contains the Kosakya XML general Utilities Classes
"""
from collections.abc import Iterator

from lxml import etree

from ut_dic.dic import Dic

from typing import Any, TypeAlias
TyDocInfo: TypeAlias = etree.DocInfo
TyTree: TypeAlias = etree._Element

TyAny = Any
TyArr = list[Any]
TyBytes = bytes
TyDic = dict[Any, Any]
TyAoB = list[TyBytes]
TyAoD = list[TyDic]
TyAoX = list[str]
TyDoA = dict[Any, TyArr]
TyDoD = dict[Any, TyDic]
TyIterAny = Iterator[Any]
TyPath = str
TyToD = tuple[TyDic, ...]

TnAny = None | Any
TnArr = None | TyArr
TnBool = None | bool
TnDic = None | TyDic


class Xml2Dic:
    """
    XML to Dictionary String Class
    Convert xml to dic, using lxml xml processing library.
    see http://lxml.de/
    """
    @staticmethod
    def init_dic(tree: TyTree) -> TyDic:
        """
        initialise tree dictionary
        """
        if tree.tag is etree.Comment:
            return {}
        if tree.attrib:
            dic: TyDic = {}
            dic[tree.tag] = {}
        else:
            if tree.tag:
                dic = {}
                dic[tree.tag] = None
            else:
                dic = {}
        return dic

    # @staticmethod
    # def normalize_value(dic):
    #     """
    #     show normalized dictionary
    #     """
    #     dic_new = {}
    #     for k, v in dic.items():
    #         if len(v) == 1:
    #             dic_new[k] = v[0]
    #         else:
    #             dic_new[k] = v
    #     return dic_new

    @classmethod
    def sh_dic(cls, iterator: Iterator[TyDic]) -> TyDic:
        """
        show children dictionary for children map
        """
        _dic: TyDoA = {}
        for c_tree in iterator:
            for k, v in c_tree.items():
                if k not in _dic:
                    _dic[k] = []
                _dic[k].append(v)

        # if value is list with 1 element use element as value
        _dic_normalized: TyDic = Dic.normalize_values(_dic)
        return _dic_normalized

    @staticmethod
    def update_with_attribs(dic: TyDic, tree: TyTree) -> None:
        """
        update dictionary with attributes dictionary
        """
        attribs = {f'@{k}': v for k, v in tree.attrib.items()}
        dic[tree.tag].update(attribs)

    @staticmethod
    def update_with_text(dic: TyDic, tree: TyTree, children) -> TyDic:
        """
        update dictionary with text
        """
        text: str = tree.text.strip()
        if children or tree.attrib:
            if text:
                dic[tree.tag]['#text'] = text
        else:
            dic[tree.tag] = text
        return dic

    @classmethod
    def update(cls, dic: TyDic, tree: TyTree, children) -> TyDic:
        """
        update dictionary
        """
        if tree.tag is etree.Comment:
            return dic
        if tree.attrib:
            cls.update_with_attribs(dic, tree)
        if tree.text:
            cls.update_with_text(dic, tree, children)
        return dic

    @staticmethod
    def map_seq(fnc, lst) -> TyArr:
        """
        apply function on list using list comprehension

        :param fnc: function
        :param List lst: List
        :return List: List of all elements of lst modified by fnc
        """
        return [fnc(entry) for entry in lst]

    @classmethod
    def tree2dic(cls, tree: TyTree) -> TyDic:
        """
        migrate tree to dictionary
        """
        children: TyArr = list(tree)
        if len(children) > 0:
            iod: Iterator[TyDic] = map(cls.tree2dic, children)
            dic: TyDic = {}
            if tree.tag is not etree.Comment:
                dic[tree.tag] = cls.sh_dic(iod)
        else:
            dic = cls.init_dic(tree)
        cls.update(dic, tree, children)
        return dic

    @classmethod
    def mig(cls, string: str) -> TyDic:
        """
        migrate xml string to etree and etree to dictionary
        """
        tree: TyTree = etree.fromstring(string)
        # Iterate through all XML elements
        for elem in tree.getiterator():
            # Skip comments and processing instructions,
            # because they do not have names
            if not (
                isinstance(elem, etree._Comment)
                or isinstance(elem, etree._ProcessingInstruction)
            ):
                # Remove a namespace URI in the element's name
                elem.tag = etree.QName(elem).localname
        etree.cleanup_namespaces(tree)
        dic = cls.tree2dic(tree)
        return {tree.tag: dic}



# coding=utf-8
"""
Koskya XML general Utilities Module
contains the Kosakya XML general Utilities Classes
"""

from lxml import etree

from typing import Any, TypeAlias
TyDocInfo: TypeAlias = etree.DocInfo
TyArr = list[Any]
TyBytes = bytes
TyAoB = list[TyBytes]
TyAoX = list[str]
TnBytes = None | TyBytes
TnDocInfo = None | TyDocInfo


class AoX:
    """
    Array of Xml Objects
    """
    @staticmethod
    def to_aob(aox: TyAoX, docinfo: TnDocInfo = None) -> TyAoB:
        aob: TyAoB = []
        if docinfo is None:
            encoding = None
        else:
            encoding = docinfo.encoding
        for item in aox:
            b_xmlstr: TyBytes = etree.tostring(
                         item,
                         xml_declaration=False,
                         with_comments=True,
                         encoding=encoding,
                         pretty_print=True)
            aob.append(b_xmlstr)
        return aob

    @classmethod
    def to_ByteStr(
            cls, aox: TyAoX,
            b_header: TnBytes = None,
            b_footer: TnBytes = None,
            docinfo: TnDocInfo = None) -> TyBytes:
        aob: TyAoB = cls.to_aob(aox, docinfo)
        return b'\n'.join(aob)



# coding=utf-8
"""
Koskya XML general Utilities Module
contains the Kosakya XML general Utilities Classes
"""

from ut_app.ut_xml import Xml2Dic

from typing import Any
TyAny = Any
TyBytes = bytes
TyAoB = list[TyBytes]
TyDic = dict[Any, Any]
TyStr = str


class AoB:
    """
    Array of Bytes
    """
    @staticmethod
    def to_Bytes(aob: TyAoB) -> TyBytes:
        return b'\n'.join(aob)

    @staticmethod
    def to_String(aob: TyAoB) -> TyStr:
        return '\n'.join(map(str, aob))

#   @classmethod
#   def to_Dic(cls, aob: TyAoB) -> TyDic:
#       _bytes: TyBytes = cls.to_Bytes(aob)
#       return Xml2Dic.mig(_bytes)

    @classmethod
    def to_Dic(cls, aob: TyAoB) -> TyDic:
        _str: str = cls.to_String(aob)
        _dic: TyDic = Xml2Dic.mig(_str)
        return _dic



# coding=utf-8
"""
Koskya XML general Utilities Module
contains the Kosakya XML general Utilities Classes
"""
from collections.abc import Iterator

from lxml import etree
import xmltodict

from ut_dic.dic import Dic
from ut_obj.byte import Byte
from .xml2dicapc import Xml2DicApc
from .aox import AoX
from .xml2dic import Xml2Dic

from typing import Any, TypeAlias
TyDocInfo: TypeAlias = etree.DocInfo
TyTree: TypeAlias = etree._Element

TyAny = Any
TyArr = list[Any]
TyBytes = bytes
TyDic = dict[Any, Any]
TyAoB = list[TyBytes]
TyAoD = list[TyDic]
TyAoX = list[str]
TyDoA = dict[Any, TyArr]
TyDoD = dict[Any, TyDic]
TyIterAny = Iterator[Any]
TyPath = str
TyToD = tuple[TyDic, ...]

TnAny = None | Any
TnArr = None | TyArr
TnBool = None | bool
TnDic = None | TyDic


class XML:
    """
    XML Class
    """
    @staticmethod
    def write_aob(aob: TyAoB, path_: TyPath) -> None:
        with open(path_, 'wb') as fd:
            for bytes in aob:
                fd.write(bytes)

    @classmethod
    def write_aox(cls, aox: TyAoX, path: TyPath, kwargs: TyDic) -> None:
        docinfo: TyDocInfo = kwargs.get('docinfo')
        aob: TyAoB = AoX.to_aob(aox, docinfo)
        with open(path, 'wb') as fd:
            for bytes in aob:
                fd.write(bytes)

    @staticmethod
    def write_aox_bytes(aox: TyAoX, path: TyPath, kwargs) -> None:
        # def write_bytes(aox, path_, kwargs) -> None:
        # aox, path_, b_header=None, b_footer=None, docinfo=None):
        b_header: TyBytes = kwargs.get('b_header')
        b_footer: TyBytes = kwargs.get('b_footer')
        docinfo: TyDocInfo = kwargs.get('docinfo')
        with open(path, 'wb') as fd:
            if b_header is not None:
                fd.write(b_header)
            for item in aox:
                out_xml = etree.tostring(
                            item,
                            xml_declaration=False,
                            with_comments=True,
                            encoding=docinfo.encoding,
                            pretty_print=True)
                fd.write(out_xml)
            if b_footer is not None:
                fd.write(b_footer)

    @staticmethod
    def read_2_obj(path: TyPath, **kwargs) -> TnAny:
        mode: str = kwargs.get('mode', 'rb')
        keys: TnArr = kwargs.get('keys')
        with open(path, mode) as fd:
            _bytes: TyBytes = fd.read()
            _str = Byte.replace_by_dic(_bytes, **kwargs)
            dic: TyDic = Xml2Dic.mig(_str)
            if keys is None:
                return None
            return Dic.get(dic, keys)

    @staticmethod
    def read_2_dic_with_xml2dicapc_mig(path, **kwargs) -> TyAny:
        mode = kwargs.get('mode', 'rb')
        with open(path, mode) as fd:
            _bytes = fd.read()
            return Xml2DicApc.mig(_bytes, **kwargs)

    @staticmethod
    def read_2_dic_with_xmltodict(path, **kwargs) -> TyAny:
        mode: str = kwargs.get('mode', 'rb')
        with open(path, mode) as fd:
            _bytes: TyBytes = fd.read()
            return xmltodict.parse(_bytes, **kwargs)

    @staticmethod
    def sh_root_docinfo(path: TyPath):
        tree: TyTree = etree.parse(path)
        docinfo: TyDocInfo = tree.docinfo
        root = tree.getroot()
        return root, docinfo



# coding=utf-8
"""
Koskya XML general Utilities Module
contains the Kosakya XML general Utilities Classes
"""
from collections.abc import Iterator

from lxml import etree

from typing import Any, TypeAlias
TyDocInfo: TypeAlias = etree.DocInfo
TyTree: TypeAlias = etree._Element

TyAny = Any
TyArr = list[Any]
TyBytes = bytes
TyDic = dict[Any, Any]
TyAoB = list[TyBytes]
TyAoD = list[TyDic]
TyAoX = list[str]
TyDoA = dict[Any, TyArr]
TyDoD = dict[Any, TyDic]
TyIterAny = Iterator[Any]
TyPath = str
TyToD = tuple[TyDic, ...]

TnAny = None | Any
TnArr = None | TyArr
TnBool = None | bool
TnDic = None | TyDic


class Xml2Dic2:

    """
    There is no one-true-mapping from an XML to a Python dict;
    one is a node tree, the other is a hash map,
    it's just an "apples and something-else comparison",
    so you'll have to make design decisions for yourself,
    considering what you want.

    The link by Sreehari has a solution that does a decent job of
    converting an lxml node to a Python dict, but:

    it requires lxml, which is fine, but I like standard modules
    when they do the job it doesn't capture attributes.
    I've taken that code and converted it work with Python's standard
    xml.ElementTree module/class, and it handles attributes in its own way.

    When I run this code against your sample, I get the following dict:

    {'fees': [{'@attribs': {'mail_retail': 'MAIL', 'member_group': '00400F'},
               'admin_fee': '0.76',
               'processing_fee': '1.83'},
              {'@attribs': {'mail_retail': 'RETAIL', 'member_group': '00400F'},
               'admin_fee': '1.335',
               'processing_fee': '1.645'},
              {'@attribs': {'mail_retail': 'MAIL', 'member_group': '00460G'},
               'admin_fee': '0.88',
           'processing_fee': '1.18'}]}
    Notice the @attribs key, that's how I decided attributes should be stored.
    If you need something else, you can modify it to your liking:
    """

    @classmethod
    def node_2_dic(cls, node) -> TyDic:
        # def elem2dict(cls, node) -> TyDic:
        """
        Convert an xml.ElementTree node tree into a dict.
        """
        result: TyDic = {}

        for element in node:
            key = element.tag
            if '}' in key:
                # Remove namespace prefix
                key = key.split('}')[1]
            if node.attrib:
                result['@attribs'] = dict(node.items())
            # Process element as tree element if the inner XML contains
            # non-whitespace content
            if element.text and element.text.strip():
                value = element.text
            else:
                value = cls.node_2_dic(element)
            # Check if a node with this name at this depth was already found
            if key in result:
                if not isinstance(result[key], list):
                    # We've seen it before, but only once, we need to
                    # convert it to a list
                    tempvalue = result[key].copy()
                    result[key] = [tempvalue, value]
                else:
                    # We've seen it at least once, it's already a list,
                    # just append the node's inner XML
                    result[key].append(value)
            else:
                # First time we've seen it
                result[key] = value
        return result

    @classmethod
    def mig(cls, path: TyPath) -> TyDic:
        from xml.etree import ElementTree as ET
        root = ET.parse('input.xml').getroot()
        return cls.node_2_dic(root)
